import asyncio
import datetime
import importlib.metadata
import logging
import os
import pathlib
import re
import sys

import niobot

try:
    import click
    import httpx
except ImportError:
    print("Missing CLI dependencies. Did you install CLI extras?", file=sys.stderr)
    sys.exit(1)
import subprocess

logger = logging.getLogger("cli")

DEFAULT_BOT_TEMPLATE = """#!/usr/bin/env python3
\"\"\"
Generated by NioBot's setup script at {timestamp} on build {version_info}. Feel free to edit this file!
\"\"\"
import niobot
import time


bot = niobot.NioBot(
    "{homeserver}",
    "{user_id}",
    "{device_id}",
    store_path="{store_path}",
    command_prefix="{prefix}",
    owner_id="{owner_id}",
)


@bot.command()
async def ping(ctx: niobot.Context):
    \"\"\"Shows the latency between the bot and the homeserver, in milliseconds.\"\"\"
    server_timestamp_seconds = ctx.message.server_timestamp / 1000
    latency = time.time() - server_timestamp_seconds
    await ctx.reply(f"Pong! {{latency * 1000:.2f}}ms")


@bot.command()
async def info(ctx: niobot.Context):
    \"\"\"Shows information about the currently running instance.\"\"\"
    await ctx.reply(f"Bot owner: {{ctx.client.owner_id}}\\n"
                    f"Bot user ID: {{ctx.client.user_id}}\\n"
                    f"Bot homeserver: {{ctx.client.homeserver}}\\n"
                    f"Bot command prefix: {{ctx.client.command_prefix}}\\n"
                    f"Bot command count: {{len(ctx.client.commands)}}\\n"
                    f"Bot module count: {{len(ctx.client.modules)}}\\n"
                    f"Bot uptime: {{time.time() - ctx.client.start_time:,.0f}} seconds")

bot.run(password="{password}")
"""


@click.group()
@click.option(
    "--log-level", "-L", type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]), default="WARNING"
)
@click.pass_context
def cli_root(ctx, log_level: str):
    """CLI utilities for nio-bot."""
    logging.basicConfig(
        level=getattr(logging, log_level),
    )

    try:
        from .__version__ import __version__, __version_tuple__
    except ImportError:
        logger.warning("Failed to import version metadata.")
        __version__ = "unknown"
        __version_tuple__ = (0, 0, "unknown", "gunknown.d19991231")
    else:
        logger.debug("Version: %s", __version__)

    ctx.obj = {"timestamp": datetime.datetime.now(), "version_info": __version__, "version_tuple": __version_tuple__}


@cli_root.command()
@click.option("--no-colour", "--no-color", "-C", is_flag=True, default=False)
@click.pass_context
def version(ctx, no_colour: bool):
    """Shows version information."""
    logger.info("Gathering version info...")
    import platform

    from nio.crypto import ENCRYPTION_ENABLED

    logger.debug("Attempting to resolve matrix-nio via importlib...")
    nio_version = None
    try:
        nio_version = importlib.metadata.version("matrix-nio")
    except importlib.metadata.PackageNotFoundError:
        logger.critical("Failed to resolve nio version information via importlib.")

    if not nio_version:
        logger.debug("Attempting to resolve matrix-nio version information via pip...")
        command = ["pip", "show", "matrix-nio"]
        result = subprocess.run(command, capture_output=True, text=True)
        if result.returncode != 0:
            logger.critical("Failed to resolve nio version information via pip.")
            nio_version = "0.0.0"
        else:
            logger.debug("Successfully resolved nio version information via pip.")
            nio_version = result.stdout.splitlines()[1].split(": ")[1]

    t = ctx.obj["version_tuple"]
    if len(t) > 3:
        t3 = t[3] or "gN/A.d%s" % (datetime.datetime.now().strftime("%Y%m%d"))
        try:
            t3_commit, t3_date_raw = t3.split(".", 1)
        except ValueError:
            t3_commit = t3
            t3_date = datetime.datetime.now()
            t3_date_raw = t3_date.strftime("%Y%m%d")
        t3_date = datetime.datetime.strptime(t3_date_raw[1:], "%Y%m%d")
    else:
        t3_commit = "<release>"
        mtime = pathlib.Path(__file__).stat().st_mtime
        t3_date = datetime.datetime.fromtimestamp(mtime)

    bot_version_deep = {"version": "%d.%d" % (t[0], t[1]), "build": t[2], "commit": t3_commit, "date": t3_date}

    bot_version = "%s (Version %s, build %s, commit %r, built %r)" % (
        ctx.obj["version_info"],
        bot_version_deep["version"],
        bot_version_deep["build"],
        bot_version_deep["commit"][1:],
        bot_version_deep["date"].strftime("%d/%m/%Y"),
    )

    _os = platform.platform()
    if hasattr(platform, "freedesktop_os_release"):
        try:
            _os_info = platform.freedesktop_os_release()
        except OSError:
            pass
        else:
            _os += " ({0}/{1} - {2})".format(
                _os_info.get("NAME", "Unknown"),
                _os_info.get("VERSION", "Unknown"),
                _os_info.get("PRETTY_NAME", "Unknown"),
            )

    lines = [
        ["NioBot version", bot_version, lambda x: True],
        ["matrix-nio version", nio_version, lambda x: x.startswith(("0.20", "0.21"))],
        ["Python version", platform.python_version(), lambda x: x.split(".")[0] == "3" and int(x.split(".")[1]) >= 9],
        ["Python implementation", platform.python_implementation(), lambda x: x == "CPython"],
        ["Operating System", _os, lambda val: val.startswith(("Windows", "Linux"))],
        ["Architecture", platform.machine(), lambda x: x == "x86_64"],
        ["OLM Installed", "Yes" if ENCRYPTION_ENABLED else "No", lambda x: x != "No"],
    ]

    _docker_path = pathlib.Path("/.dockerenv")
    if _docker_path.is_file():
        lines.append(["Running in Docker", "Probably (/.dockerenv exists)", lambda _: True])

    click.echo()
    for line in lines:
        if no_colour:
            click.echo("%s: %s" % tuple(line)[:2])
        else:
            click.echo(
                click.style(line[0], fg="cyan") + ": " + click.style(line[1], fg="green" if line[2](line[1]) else "red")
            )
    click.echo()


@cli_root.command(name="test-homeserver")
@click.argument("homeserver")
def test_homeserver(homeserver: str):
    """Walks through resolving and testing a given homeserver."""
    import urllib.parse

    import httpx

    if homeserver.startswith("@"):
        _, homeserver = homeserver.split(":", 1)

    logger.debug("Parsing given input %r as a URL...", homeserver)
    parsed = urllib.parse.urlparse(homeserver)
    if not parsed.scheme:
        logger.info("No scheme found, assuming HTTPS.")
        parsed = urllib.parse.urlparse("https://" + homeserver)

    if not parsed.netloc:
        logger.critical("No netloc found, cannot continue.")
        logger.critical("URI parsed to: %r", parsed)
        return

    logger.debug("Attempting to resolve %r...", parsed.netloc)
    logger.info("Trying well-known of %r...", parsed.netloc)
    base_url = None
    try:
        response = httpx.get("https://%s/.well-known/matrix/client" % parsed.netloc, timeout=30)
    except httpx.HTTPError as e:
        logger.critical("Failed to get well-known: %r", e)
        return
    else:
        logger.debug("Got response: %r", response)
        cors = response.headers.get("Access-Control-Allow-Origin", None)
        if not cors:
            logger.warning("No CORS header found. This homeserver will be unusable in web-based clients!")
        elif cors != "*":
            logger.warning(
                "CORS header found, but not set to wildcard. "
                "This homeserver will be unusable in most web-based clients!"
            )

        if response.status_code != 404:
            if response.status_code != 200 or len(response.content or b"") == 0:
                logger.critical("Well-known returned non-404, but no content. Failing.")
                return
            else:
                data = response.json()
                logger.debug("Well-known data: %r", data)
                if "m.homeserver" not in data:
                    logger.critical("Well-known data does not contain m.homeserver. Failing.")
                    return
                base_url = urllib.parse.urlparse(data["m.homeserver"]["base_url"])
                logger.debug("Parsed base URL: %r", base_url)
                if not base_url.scheme or not base_url.netloc:
                    logger.critical("Base URL does not contain a scheme. Invalid URL? Failing.")
                    return

    if not base_url:
        logger.info("No well-known found. Assuming %r as homeserver.", parsed.netloc)
        base_url = urllib.parse.urlparse("https://" + parsed.netloc)

    base_url = base_url.geturl()
    logger.info("Using %r as homeserver.", base_url)
    logger.info("Validating homeserver...")
    try:
        response = httpx.get(base_url + "/_matrix/client/versions", timeout=30)
    except httpx.HTTPError as e:
        logger.critical("Failed to get versions: %r", e)
        return
    else:
        data = response.json()
        logger.debug("Got response: %r", data)
        if "versions" not in data:
            logger.critical("Versions response does not contain versions. Failing.")
            return
        if "unstable_features" not in data:
            logger.warning("Versions response does not contain unstable_features. This may be bad.")

        logger.info("Homeserver validated.")
        click.secho("Homeserver validated.", fg="green")
        click.secho("%s -> %s" % (click.style(parsed.netloc, dim=True), click.style(base_url, fg="green")))


@cli_root.command(name="get-access-token")
@click.option("--username", "-U", default=None, help="The username part (without @)")
@click.option("--password", "-P", default=None, help="The password to use (will be prompted if not given)")
@click.option("--homeserver", "-H", default=None, help="The homeserver to use (will be prompted if not given)")
@click.option(
    "--device-id",
    "-D",
    "--session",
    "--session-id",
    default=None,
    help="The device ID to use (will be prompted if not given)",
)
@click.pass_context
def get_access_token(ctx, username: str, password: str, homeserver: str, device_id: str):
    """Fetches your access token from your homeserver."""
    if not username:
        username = ""

    while not re.match(r"@[\w\-.]{1,235}:[0-9A-Za-z\-_.]{3,253}", username):
        username = click.prompt("User ID (@username:homeserver.tld)")
        _, homeserver = username.split(":", 1)

    if not homeserver:
        _, homeserver = username.split(":", 1)

    if not password:
        password = click.prompt("Password (will not echo)", hide_input=True)

    if not homeserver:
        homeserver = click.prompt("Homeserver URL")

    if not device_id:
        device_id = click.prompt(
            "Device ID (a memorable display name for this login, such as 'bot-production')", default=os.uname().nodename
        )

    click.secho("Resolving homeserver... ", fg="cyan", nl=False)
    try:
        homeserver = asyncio.run(niobot.resolve_homeserver(homeserver))
    except ConnectionError:
        click.secho("Failed!", fg="red")
    else:
        click.secho("OK", fg="green")

    click.secho("Getting access token... ", fg="cyan", nl=False)
    status_code = None
    try:
        response = httpx.post(
            homeserver + "/_matrix/client/r0/login",
            json={
                "type": "m.login.password",
                "identifier": {"type": "m.id.user", "user": username},
                "password": password,
                "device_id": device_id,
                "initial_device_display_name": device_id,
            },
        )
        status_code = response.status_code
        if status_code == 429:
            click.secho("Failed!", fg="red", nl=False)
            click.secho(" (Rate limited for {:.0f} seconds)".format(response.json()["retry_after_ms"] / 1000), bg="red")
            return
        response.raise_for_status()
    except httpx.HTTPError as e:
        click.secho("Failed!", fg="red", nl=False)
        click.secho(" (%s)" % status_code or str(e), bg="red")
    else:
        click.secho("OK", fg="green")
        click.secho("Access token: %s" % response.json()["access_token"], fg="green")


@cli_root.group()
def new():
    """Create a new file from a template."""


@new.command()
@click.argument("path", type=click.Path())
@click.option("--password", "-P", prompt=True, hide_input=True)
@click.option("--user-id", "-U", prompt=True)
@click.option("--homeserver", "-H", prompt=True)
@click.option("--device-id", "-D", prompt=True)
@click.option("--store-path", "-S", prompt=True)
@click.option("--prefix", "-P", prompt=True)
@click.option("--owner-id", "-O", prompt=True)
@click.pass_context
def bot(
    ctx,
    path: str,
    password: str,
    homeserver: str,
    user_id: str,
    device_id: str,
    store_path: str,
    prefix: str,
    owner_id: str,
):
    """Creates a new bot with two basic commands from the template. Easy quickstart."""
    with open(path, "w") as f:
        f.write(
            DEFAULT_BOT_TEMPLATE.format(
                password=password,
                homeserver=homeserver,
                user_id=user_id,
                device_id=device_id,
                store_path=store_path,
                prefix=prefix,
                owner_id=owner_id,
                timestamp=datetime.datetime.utcnow().isoformat(),
                version_info=ctx.obj["version_info"],
            )
        )
    click.echo("Created bot file at %r" % path)


if __name__ == "__main__":
    cli_root()



# WARNING
# This is INSECURE
# This code will store your USER ACCESS TOKEN in PLAIN TEXT JSON.
# This means that if someone were to get access to the file, they would have full access to whatever account you're
# using with niobot. They would not need your password, and would have full control.
# Please consider alternative forms of storage, such as a database, or a more secure file format.

import nio  # remember, Nio is the underlying library that niobot uses.
import niobot
from pathlib import Path
import json

CREDENTIALS = Path("./credentials.json")
if CREDENTIALS.exists():
    print("Loading existing credentials...")
    with CREDENTIALS.open() as f:
        credentials = json.load(f)
    homeserver = credentials["homeserver"]
    user_id = credentials["user_id"]
    access_token = credentials["access_token"]
    password = None
else:
    import getpass
    access_token = None
    print("No existing credentials. Please enter:")
    homeserver = input("Homeserver URL: ")
    user_id = input("User ID: ")
    password = getpass.getpass("Password (no echo): ")

bot = niobot.NioBot(homeserver, user_id, command_prefix="$")


@bot.on_event("ready")
async def on_ready(first_sync_result: nio.SyncResponse):
    print("Logged in as %r!" % bot.user_id)
    print("Saving credentials...")
    creds = {
        "homeserver": homeserver,
        "user_id": user_id,
        "access_token": bot.access_token
    }
    with CREDENTIALS.open("w") as file:
        json.dump(creds, file)
    print("Dumped!")

bot.run(access_token=access_token, password=password)
# If password is None, it will use access_token, and vice versa.



import niobot
import aiohttp
import io
import shutil

bot = niobot.NioBot(
    "https://matrix.org",
    "@user:matrix.org",
    command_prefix="£"
)


@bot.command("upload")
async def upload_random_image(ctx: niobot.Context):
    """Uploads a random inspirational quote!"""
    if not shutil.which("ffprobe"):
        await ctx.respond("ffprobe is not installed! This command requires ffprobe (part of ffmpeg) to work.")
        return

    async with aiohttp.ClientSession() as session:
        async with session.get("https://inspirobot.me/api?generate=true") as resp:
            if resp.status != 200:
                await ctx.respond("Failed to fetch image!")
                return
            url = await resp.text()
            async with session.get(url) as resp:
                if resp.status != 200:
                    await ctx.respond("Failed to fetch image!")
                    return
                with open("./tmp-image.png", "w+") as file:
                    file.write(await resp.read())
                attachment = await niobot.MediaAttachment.from_file("./tmp-image.png")
                await ctx.respond("Here's your image!", file=attachment)


bot.run(password="my_password")



"""
This utility modules contains a handful of simple off-the-shelf parser for some basic python types.
"""

import re
import typing
import urllib.parse as urllib
from collections import namedtuple

import nio

from ..exceptions import CommandParserError

if typing.TYPE_CHECKING:
    from ..commands import Argument
    from ..context import Context


__all__ = (
    "boolean_parser",
    "float_parser",
    "integer_parser",
    "json_parser",
    "event_parser",
    "room_parser",
    "BUILTIN_MAPPING",
    "MATRIX_TO_REGEX",
    "MatrixToLink",
    "MatrixMXCUrl",
    "matrix_to_parser",
    "mxc_url_parser",
)
MATRIX_TO_REGEX = re.compile(
    r"(http(s)?://)?matrix\.to/#/(?P<room_id>[^/]+)(/(?P<event_id>[^/?&#]+))?(?P<qs>([&?](via=[^&]+))*)?",
)
MatrixToLink = namedtuple("MatrixToLink", ("room", "event", "query"), defaults=(None, None, None))
MatrixMXCUrl = namedtuple("MatrixMXCUrl", ("server", "media_id"), defaults=(None, None))


def boolean_parser(_: "Context", __, value: str) -> bool:
    """
    Converts a given string into a boolean. Value is lower-cased before being parsed.

    The following resolves to true:
    * 1, y, yes, true, on

    The following resolves to false:
    * 0, n, no, false, off

    The following will raise a command argument error: anything else

    :return: The parsed boolean
    """
    value = value.lower()
    if value in ("1", "y", "yes", "true", "on"):
        return True
    if value in ("0", "n", "no", "false", "off"):
        return False
    raise CommandParserError(f"Invalid boolean value: {value}. Should be a sensible value, such as 1, yes, false.")


def float_parser(_: "Context", __: "Argument", value: str) -> float:
    """
    Converts a given string into a floating point number.

    :return: A parsed boolean
    :raises CommandParserError: if the value is not a valid number.
    """
    try:
        return float(value)
    except ValueError:
        raise CommandParserError(f"Invalid float value: {value}. Should be a number.")


def integer_parser(
    allow_floats: bool = False, base: int = 10
) -> typing.Callable[["Context", "Argument", str], typing.Union[int, float]]:
    """
    Converts a given value into an integer, or a float if allowed.

    :param allow_floats: Whether to allow parsing for floating numbers (decimals). Defaults to False.
    :param base: The base to parse (defaults to base 10, denary)
    :return: The parsed number.
    :raises CommandParserError: if the value is not a valid number.
    """

    def __parser(_, __, v) -> typing.Union[int, float]:
        try:
            return int(v, base)
        except ValueError as e:
            if allow_floats:
                try:
                    return float_parser(_, __, v)
                except CommandParserError as e2:
                    e = e2
            raise CommandParserError(f"Invalid integer value: {v}. Should be a number.", exception=e)

    return __parser


def json_parser(
    _: "Context", __: "Argument", value: str
) -> typing.Union[list, dict, str, int, float, type(None), bool]:
    """
    Converts a given string into a JSON object.

    !!! note "Performance boost"
        If you want this to be fast, you should install orjson. It is a drop-in replacement for the standard library.
        While the parser will still work without it, it may be slower, especially for larger payloads.

    :return: The parsed JSON object
    :raises CommandParserError: if the value is not a valid JSON object.
    """
    try:
        import orjson as json
    except ImportError:
        # Orjson is not installed.
        import json

    try:
        return json.loads(value)
    except json.JSONDecodeError as e:
        raise CommandParserError(f"Invalid JSON value: {value}. Should be a valid JSON object.", exception=e)


async def room_parser(ctx: "Context", arg: "Argument", value: str) -> nio.MatrixRoom:
    """
    Parses a room ID, alias, or matrix.to link into a MatrixRoom object.

    ??? warning "This parser is async"
        This parser is async, and should be awaited when used manually.

    :return: The MatrixRoom object
    """
    if value.startswith("!"):
        # Room ID
        room = ctx.client.rooms.get(value)
    elif value.startswith("#"):
        # Room alias
        # Attempt to find the room in the cache, before requesting it from the server
        for room in ctx.client.rooms.values():
            room: nio.MatrixRoom
            if value == room.canonical_alias:
                break
        else:
            room: nio.RoomResolveAliasResponse | nio.RoomResolveAliasError = await ctx.client.room_resolve_alias(value)
            if isinstance(room, nio.RoomResolveAliasError):
                raise CommandParserError(f"Invalid room alias: {value}.", response=room)
            room = ctx.client.rooms.get(room.room_id)
    elif m := MATRIX_TO_REGEX.match(value):
        # matrix.to link
        groups = m.groupdict()
        if "room_id" not in groups:
            raise CommandParserError(f"Invalid matrix.to link: {value} (no room).")
        room_id = urllib.unquote(groups["room_id"])
        room = ctx.client.rooms.get(groups["room_id"])
    else:
        raise CommandParserError(f"Invalid room ID, alias, or matrix.to link: {value!r}.")

    if room is None:
        raise CommandParserError(f"No room with that ID, alias, or matrix.to link found.")
    return room


def event_parser(
    event_type: str = None,
) -> typing.Callable[["Context", "Argument", str], typing.Coroutine[typing.Any, typing.Any, nio.Event]]:
    """
    Parses an event reference from either its ID, or matrix.to link.

    :param event_type: The event type to expect (such as m.room.message). If None, any event type is allowed.
    :return: The actual internal (async) parser.
    """

    async def internal(ctx: "Context", _, value: str) -> nio.Event:
        room_id = ctx.room.room_id
        if m := MATRIX_TO_REGEX.match(value):
            # matrix.to link
            groups = m.groupdict()
            if "room_id" not in groups:
                raise CommandParserError(f"Invalid matrix.to link: {value} (no room).")
            if "event_id" not in groups:
                raise CommandParserError(f"Invalid matrix.to link: {value} (expected an event).")
            value = groups["event_id"]
            room_id = groups["room_id"]
            value = urllib.unquote(value)
            room_id = urllib.unquote(room_id)

        if value.startswith("$"):
            # from raw ID
            event: nio.RoomGetEventResponse | nio.RoomGetEventError = await ctx.client.room_get_event(room_id, value)
            if not isinstance(event, nio.RoomGetEventResponse):
                raise CommandParserError(f"Invalid event ID: {value}.", response=event)
            event: nio.Event = event.event
            if event_type is not None and event_type != event.source.get("type"):
                raise CommandParserError(
                    f"Invalid event ID: {value} (expected {event_type}, got {event.source.get('type')})."
                )
            return event
        else:
            raise CommandParserError(f"Invalid event ID or matrix.to link: {value!r}.")

    return internal


def matrix_to_parser(
    require_room: bool = True, require_event: bool = False, allow_user_as_room: bool = True
) -> typing.Callable[["Context", "Argument", str], typing.Coroutine[typing.Any, typing.Any, MatrixToLink],]:
    """
    Advanced multi-functional parser for matrix.to links.

    This parser takes an https://matrix.to link, and breaks it down into room-id and event-id where applicable. It then
    returns all of these as a namedtuple.

    :param require_room: Whether to require the room part of this url to be present
    :param require_event: Whether to require the event part of this url to be present
    :param allow_user_as_room: Whether to allow user links as room links
    :return: The actual internal (async) parser.
    """

    async def internal(ctx: "Context", _, value: str) -> MatrixToLink:
        if m := MATRIX_TO_REGEX.match(value):
            # matrix.to link
            groups = m.groupdict()
            event_id = groups.get("event_id", "")
            room_id = groups.get("room_id", "")
            event_id = urllib.unquote(event_id)
            room_id = urllib.unquote(room_id)

            if require_room and not room_id:
                raise CommandParserError(f"Invalid matrix.to link: {value} (no room).")
            if require_event and not event_id:
                raise CommandParserError(f"Invalid matrix.to link: {value} (no event).")

            if room_id.startswith("@") and not allow_user_as_room:
                raise CommandParserError(f"Invalid matrix.to link: {value} (expected room, got user).")

            if room_id.startswith("@"):
                room = await ctx.client.get_dm_room(room_id)
            else:
                room = ctx.client.rooms.get(room_id)

            if room is None:
                raise CommandParserError(f"No room with that ID, alias, or matrix.to link found.")

            if event_id:
                event: nio.RoomGetEventResponse | nio.RoomGetEventError = await ctx.client.room_get_event(
                    room_id, event_id
                )
                if not isinstance(event, nio.RoomGetEventResponse):
                    raise CommandParserError(f"Invalid event ID: {event_id}.", response=event)
                event: nio.Event = event.event
            else:
                event: None = None
            return MatrixToLink(room, event, groups.get("qs"))
        else:
            raise CommandParserError(f"Invalid matrix.to link: {value!r}.")

    return internal


def mxc_url_parser(_, __, value: str) -> MatrixMXCUrl:
    """
    Parses an MXC URL. Basically just a validator
    """
    if not value.startswith("mxc://"):
        raise CommandParserError(f"Invalid MXC URL: {value!r}.")
    parsed = urllib.urlparse(value, "mxc://", allow_fragments=False)
    if not parsed.path:
        raise CommandParserError(f"Invalid MXC URL: {value!r} (missing media ID).")
    if parsed.path == "/":
        raise CommandParserError(f"Invalid MXC URL: {value!r} (no media ID).")
    if not parsed.netloc:
        raise CommandParserError(f"Invalid MXC URL: {value!r} (no server).")

    return MatrixMXCUrl(parsed.netloc, parsed.path[1:])


BUILTIN_MAPPING = {
    bool: boolean_parser,
    float: float_parser,
    int: integer_parser,
    list: json_parser,
    dict: json_parser,
    nio.RoomMessageText: event_parser("m.room.message"),
    nio.Event: event_parser(),
    nio.MatrixRoom: room_parser,
    MatrixToLink: matrix_to_parser(require_event=True),
}



import niobot
import time


bot = niobot.NioBot(
    "https://matrix.org",
    "@my_user_id:matrix.org",  # remember, you can use any account here. It doesn't have to be special.
    "my_bot_name",
    store_path="/tmp",  # store state in a temporary directory, we don't really need it for these examples.
    command_prefix="!",
    owner_id="@my_user_id:matrix.org"  # this is the user ID of the bot's owner. It's used for the owner-only commands.
)


@bot.command()
async def ping(ctx: niobot.Context):
    """Shows the latency between the bot and the homeserver, in milliseconds."""
    server_timestamp_seconds = ctx.message.server_timestamp / 1000  # convert from milliseconds to seconds
    latency = time.time() - server_timestamp_seconds  # the time between now and the server timestamp
    await ctx.reply(f"Pong! {latency * 1000:.2f}ms")  # convert to ms, and reply to the user


@bot.command()
async def info(ctx: niobot.Context):
    """Shows information about the currently running instance."""
    await ctx.reply(f"Bot owner: {ctx.client.owner_id}\n"
                    f"Bot user ID: {ctx.client.user_id}\n"
                    f"Bot homeserver: {ctx.client.homeserver}\n"
                    f"Bot command prefix: {ctx.client.command_prefix}\n"
                    f"Bot command count: {len(ctx.client.commands)}\n"
                    f"Bot module count: {len(ctx.client.modules)}\n"
                    f"Bot uptime: {time.time() - ctx.client.start_time:,.0f} seconds")

bot.run(password="my_account_password")  # it is recommended to use an access token instead. See access_token.py



import inspect
import logging
import os
import typing

import nio

from .context import Context
from .exceptions import *

if typing.TYPE_CHECKING:
    from .client import NioBot


__all__ = ("Command", "command", "event", "Module", "Argument", "check")

_T = typing.TypeVar("_T")


class Argument:
    """
    Represents a command argument.

    ??? example
        ```py
        from niobot import NioBot, command, Argument

        bot = NioBot(...)

        @bot.command("echo")
        def echo(ctx: niobot.Context, message: str):
            await ctx.respond(message)

        bot.run(...)
        ```

    :param name: The name of the argument. Will be used to know which argument to pass to the command callback.
    :param arg_type: The type of the argument (e.g. str, int, etc. or a custom type)
    :param description: The description of the argument. Will be shown in the auto-generated help command.
    :param default: The default value of the argument
    :param required: Whether the argument is required or not. Defaults to True if default is ..., False otherwise.
    """

    def __init__(
        self,
        name: str,
        arg_type: _T,
        *,
        description: str = None,
        default: typing.Any = ...,
        required: bool = ...,
        parser: typing.Callable[["Context", "Argument", str], typing.Optional[_T]] = ...,
        **kwargs,
    ):
        log = logging.getLogger(__name__)
        self.name = name
        self.type = arg_type
        self.description = description
        self.default = default
        self.required = required
        if self.required is ...:
            self.required = default is ...
            self.default = None
        self.extra = kwargs
        self.parser = parser
        if self.parser is ...:
            from .utils import BUILTIN_MAPPING

            if self.type in BUILTIN_MAPPING:
                log.info("Using builtin parser %r for %r", self.type, self)
                self.parser = BUILTIN_MAPPING[self.type]
            else:
                for base in inspect.getmro(self.type):
                    if base in BUILTIN_MAPPING:
                        log.info("Using builtin parser %r for %s due to being a subclass", base, self.type)
                        self.parser = BUILTIN_MAPPING[base]
                        break
                else:
                    log.info("Using default parser for %s", self.type)
                    self.parser = self.internal_parser

    def __repr__(self):
        return (
            "<Argument name={0.name!r} type={0.type!r} default={0.default!r} required={0.required!r} "
            "parser={0.parser!r}>".format(self)
        )

    @staticmethod
    def internal_parser(_: Context, arg: "Argument", value: str) -> typing.Optional[_T]:
        """The default parser for the argument. Will try to convert the value to the argument type."""
        try:
            return arg.type(value)
        except ValueError:
            raise CommandParserError(f"Invalid value for argument {arg.name}: {value!r}")


class Command:
    """Represents a command.

    ??? example
        !!! note
            This example uses the `command` decorator, but you can also use the [`Command`][niobot.commands.Command]
            class directly, but you
            likely won't need to, unless you want to pass a custom command class.

            All that the `@command` decorator does is create a [`Command`][niobot.commands.Command] instance and
            add it to the bot's commands,
            while wrapping the function its decorating.

        ```py
        from niobot import NioBot, command

        bot = NioBot(...)

        @bot.command("hello")
        def hello(ctx: niobot.Context):
            await ctx.respond("Hello, %s!" % ctx.message.sender)

        bot.run(...)
        ```

    :param name: The name of the command. Will be used to invoke the command.
    :param callback: The callback to call when the command is invoked.
    :param aliases: The aliases of the command. Will also be used to invoke the command.
    :param description: The description of the command. Will be shown in the auto-generated help command.
    :param disabled:
        Whether the command is disabled or not. If disabled, the command will be hidden on the auto-generated
        help command, and will not be able to be invoked.
    :param arguments:
        A list of [`Argument`][niobot.commands.Argument] instances. Will be used to parse the arguments given to the
        command.
        `ctx` is always the first argument, regardless of what you put here.
    :param usage:
        A string representing how to use this command's arguments. Will be shown in the auto-generated help.
        Do not include the command name or your bot's prefix here, only arguments.
        For example: `usage="<message> [times]"` will show up as `[p][command] <message> [times]` in the help command.
    :param hidden:
        Whether the command is hidden or not. If hidden, the command will be always hidden on the auto-generated help.
    """

    _CTX_ARG = Argument("ctx", Context, description="The context for the command", parser=lambda ctx, *_: ctx)

    def __init__(
        self,
        name: str,
        callback: callable,
        *,
        aliases: list[str] = None,
        description: str = None,
        disabled: bool = False,
        hidden: bool = False,
        **kwargs,
    ):
        self.__runtime_id = os.urandom(16).hex()
        self.log = logging.getLogger(__name__)
        self.name = name
        self.callback = callback
        self.description = description
        self.disabled = disabled
        self.aliases = aliases or []
        self.checks = kwargs.pop("checks", [])
        if hasattr(self.callback, "__nio_checks__"):
            for check_func in self.callback.__nio_checks__.keys():
                self.checks.append(check_func)
        self.hidden = hidden
        self.usage = kwargs.pop("usage", None)
        self.module = kwargs.pop("module", None)
        self.arguments = kwargs.pop("arguments", None)
        if not self.arguments:
            if self.arguments is False:  # do not autodetect arguments
                self.arguments = []
            else:
                self.arguments = self.autodetect_args(self.callback)
        self.arguments.insert(0, self._CTX_ARG)
        self.arguments: list[Argument]

    @staticmethod
    def autodetect_args(callback) -> list[Argument]:
        """
        Attempts to auto-detect the arguments for the command, based on the callback's signature

        :param callback: The function to inspect
        :return: A list of arguments. `self`, and `ctx` are skipped.
        """
        # We need to get each parameter's type annotation, and create an Argument for it.
        # If it has a default value, assign that default value to the Argument.
        # If the parameter is `self`, ignore it.
        # If the parameter is `ctx`, use the `Context` type.
        args = []
        for n, parameter in enumerate(inspect.signature(callback).parameters.values()):
            # If it has a parent class and this is the first parameter, skip it.
            if n == 0 and parameter.name == "self":
                continue

            if parameter.name in ["ctx", "context"] or parameter.annotation is Context:
                continue

            # Disallow *args and **kwargs
            if parameter.kind in [inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD]:
                raise CommandArgumentsError("Cannot use *args or **kwargs in command callback (argument No. %d)" % n)

            if parameter.annotation is inspect.Parameter.empty:
                a = Argument(parameter.name, str, default=parameter.default)
            else:
                a = Argument(parameter.name, parameter.annotation)

            if parameter.default is not inspect.Parameter.empty:
                a.default = parameter.default
                a.required = False
            args.append(a)
        return args

    def __hash__(self):
        return hash(self.__runtime_id)

    def __eq__(self, other):
        """Checks if another command's runtime ID is the same as this one's"""
        if isinstance(other, Command):
            return self.__runtime_id == other.__runtime_id
        else:
            return False

    def __repr__(self):
        return "<Command name={0.name!r} aliases={0.aliases} disabled={0.disabled}>".format(self)

    def __str__(self):
        return self.name

    @property
    def display_usage(self) -> str:
        """Returns the usage string for this command, auto-resolved if not pre-defined"""
        if self.usage:
            return self.usage
        else:
            usage = []
            req = "<{!s}>"
            opt = "[{!s}]"
            for arg in self.arguments[1:]:
                if arg.required:
                    usage.append(req.format(arg.name))
                else:
                    usage.append(opt.format(arg.name))
            return " ".join(usage)

    async def invoke(self, ctx: Context) -> typing.Coroutine:
        """
        Invokes the current command with the given context

        :param ctx: The current context
        :raises CommandArgumentsError: Too many/few arguments, or an error parsing an argument.
        :raises CheckFailure: A check failed
        """
        from .utils import force_await

        if self.checks:
            for chk_func in self.checks:
                name = self.callback.__nio_checks__[chk_func]
                try:
                    cr = await force_await(chk_func, ctx)
                except CheckFailure:
                    raise  # re-raise existing check failures
                except Exception as e:
                    raise CheckFailure(name, exception=e) from e
                if not cr:
                    raise CheckFailure(name)

        parsed_args = []
        if len(ctx.args) > (len(self.arguments) - 1):
            raise CommandArgumentsError(f"Too many arguments given to command {self.name}")
        for index, argument in enumerate(self.arguments[1:]):
            argument: Argument

            if index >= len(ctx.args):
                if argument.required:
                    raise CommandArgumentsError(f"Missing required argument {argument.name}")
                else:
                    parsed_args.append(argument.default)
                    continue

            self.log.debug("Resolved argument %s to %r", argument.name, ctx.args[index])
            try:
                parsed_argument = await force_await(argument.parser, ctx, argument, ctx.args[index])
            except Exception as e:
                error = f"Error while parsing argument {argument.name}: {e}"
                raise CommandArgumentsError(error) from e
            parsed_args.append(parsed_argument)

        parsed_args = [ctx, *parsed_args]
        self.log.debug("Arguments to pass: %r", parsed_args)
        if self.module:
            self.log.debug("Will pass module instance")
            return self.callback(self.module, *parsed_args)
        else:
            return self.callback(*parsed_args)

    def construct_context(
        self, client: "NioBot", room: nio.MatrixRoom, src_event: nio.RoomMessageText, meta: str
    ) -> Context:
        return Context(client, room, src_event, self, invoking_string=meta)


def command(name: str = None, **kwargs) -> callable:
    """
    Allows you to register commands later on, by loading modules.

    This differs from NioBot.command() in that commands are not automatically added, you need to load them with
    bot.mount_module
    :param name: The name of the command. Defaults to function.__name__
    :param kwargs: Any key-words to pass to Command
    :return:
    """
    cls = kwargs.pop("cls", Command)

    def decorator(func):
        nonlocal name
        name = name or func.__name__
        cmd = cls(name, func, **kwargs)
        func.__nio_command__ = cmd
        return func

    return decorator


def check(
    function: typing.Callable[[Context], typing.Union[bool, typing.Coroutine[None, None, bool]]],
    name: str = None,
) -> callable:
    """
    Allows you to register checks in modules.

    ```python
    @niobot.command()
    @niobot.check(my_check_func, name="My Check")
    async def my_command(ctx: niobot.Context):
        pass
    ```

    :param function: The function to register as a check
    :param name: A human-readable name for the check. Defaults to function.__name__
    :return: The decorated function.
    """

    def decorator(command_function):
        if hasattr(command_function, "__nio_checks__"):
            command_function.__nio_checks__[function] = name or function.__name__
        else:
            command_function.__nio_checks__ = {function: name or function.__name__}
        return command_function

    decorator.internal = function
    return decorator


def event(name: str) -> callable:
    """
    Allows you to register event listeners in modules.

    :param name: the name of the event (no `on_` prefix)
    :return:
    """

    def decorator(func):
        func.__nio_event__ = {"function": func, "name": name, "_module_instance": None}
        return func

    return decorator


class Module:
    __is_nio_module__ = True

    def __init__(self, bot: "NioBot"):
        self.bot = self.client = bot
        self.log = logging.getLogger(__name__)

    def list_commands(self):
        for _, potential_command in inspect.getmembers(self):
            if hasattr(potential_command, "__nio_command__"):
                yield potential_command.__nio_command__

    def list_events(self):
        for _, potential_event in inspect.getmembers(self):
            if hasattr(potential_event, "__nio_event__"):
                yield potential_event.__nio_event__

    def _event_handler_callback(self, function):
        # Due to the fact events are less stateful than commands, we need to manually inject self for events
        async def wrapper(*args, **kwargs):
            return await function(self, *args, **kwargs)

        return wrapper

    def __setup__(self):
        """Setup function called once by NioBot.mount_module(). Mounts every command discovered."""
        for cmd in self.list_commands():
            cmd.module = self
            logging.getLogger(__name__).debug("Discovered command %r in %s.", cmd, self.__class__.__name__)
            self.bot.add_command(cmd)

        for _event in self.list_events():
            _event["_module_instance"] = self
            self.bot.add_event_listener(_event["name"], self._event_handler_callback(_event["function"]))

    def __teardown__(self):
        """Teardown function called once by NioBot.unmount_module(). Removes any command that was mounted."""
        for cmd in self.list_commands():
            self.bot.remove_command(cmd)
        del self
